import { SignJWT, jwtVerify, errors } from "jose";
import type Koa from "koa";
import Router from "koa-router";
import type { IUser, UserAccessor } from "./model/mod.ts";
import { sendError } from "./route/error_handler.ts";
import { get_setting } from "./SettingConfig.ts";
import { LoginRequestSchema, LoginResetRequestSchema } from "dbtype";

type PayloadInfo = {
	username: string;
	permission: string[];
};

export type UserState = {
	user: PayloadInfo;
};

const isUserState = (obj: object | string): obj is PayloadInfo => {
	if (typeof obj === "string") return false;
	return "username" in obj && "permission" in obj && Array.isArray((obj as { permission: unknown }).permission);
};

type RefreshPayloadInfo = { username: string };
const isRefreshToken = (obj: object | string): obj is RefreshPayloadInfo => {
	if (typeof obj === "string") return false;
	return "username" in obj && typeof (obj as { username: unknown }).username === "string";
};

const accessExpiredTime = 60 * 60 * 2; // 2 hour
async function createAccessToken(payload: PayloadInfo, secret: string) {
	return await new SignJWT(payload)
		.setProtectedHeader({ alg: "HS256" })
		.setExpirationTime("2h")
		.sign(new TextEncoder().encode(secret));
}

const refreshExpiredTime = 60 * 60 * 24 * 14; // 14 day;
async function createRefreshToken(payload: RefreshPayloadInfo, secret: string) {
	return await new SignJWT(payload)
		.setProtectedHeader({ alg: "HS256" })
		.setExpirationTime("14d")
		.sign(new TextEncoder().encode(secret));
}

class TokenExpiredError extends Error {
	constructor() {
		super("Token expired");
	}
}

async function verifyToken(token: string, secret: string) {
	try {
		const { payload } = await jwtVerify(token, new TextEncoder().encode(secret));
		return payload as PayloadInfo;
	} catch (error) {
		if (error instanceof errors.JWTExpired) {
			throw new TokenExpiredError();
		}
		throw new Error("Invalid token");
	}
}

export const accessTokenName = "access_token";
export const refreshTokenName = "refresh_token";

function setToken(ctx: Koa.Context, token_name: string, token_payload: string | null, expiredtime: number) {
	const setting = get_setting();
	if (token_payload === null && !ctx.cookies.get(token_name)) {
		return;
	}
	ctx.cookies.set(token_name, token_payload, {
		httpOnly: true,
		secure: setting.secure,
		sameSite: "strict",
		expires: new Date(Date.now() + expiredtime * 1000),
	});
}

export const createLoginMiddleware = (userController: UserAccessor) => async (ctx: Koa.Context, _next: Koa.Next) => {
	const setting = get_setting();
	const secretKey = setting.jwt_secretkey;
	const body = ctx.request.body;
	const {
		username,
		password,
	} = LoginRequestSchema.parse(body);
	// if admin login is forbidden?
	if (username === "admin" && setting.forbid_remote_admin_login) {
		return sendError(403, "forbidden remote admin login");
	}
	const user = await userController.findUser(username);
	// username not exist
	if (user === undefined) return sendError(401, "not authorized");
	// password not matched
	if (!user.password.check_password(password)) {
		return sendError(401, "not authorized");
	}
	// create token
	const userPermission = await user.get_permissions();
	const payload = await createAccessToken({
		username: user.username,
		permission: userPermission,
	}, secretKey);
	const payload2 = await createRefreshToken({
		username: user.username,
	}, secretKey);
	setToken(ctx, accessTokenName, payload, accessExpiredTime);
	setToken(ctx, refreshTokenName, payload2, refreshExpiredTime);
	ctx.body = {
		username: user.username,
		permission: userPermission,
		accessExpired: Math.floor(Date.now() / 1000) + accessExpiredTime,
	};
	console.log(`${username} logined`);
	return;
};

export const LogoutMiddleware = (ctx: Koa.Context, next: Koa.Next) => {
	const setting = get_setting();
	ctx.cookies.set(accessTokenName, null);
	ctx.cookies.set(refreshTokenName, null);
	ctx.body = {
		ok: true,
		username: "",
		permission: setting.guest,
	};
	return;
};

export const createUserMiddleWare =
	(userController: UserAccessor) => async (ctx: Koa.ParameterizedContext<UserState>, next: Koa.Next) => {
		const refreshToken = refreshTokenHandler(userController);
		const setting = get_setting();
		const setGuest = async () => {
			setToken(ctx, accessTokenName, null, 0);
			setToken(ctx, refreshTokenName, null, 0);
			ctx.state.user = { username: "", permission: setting.guest };
			return await next();
		};
		return await refreshToken(ctx, setGuest, next);
	};

const refreshTokenHandler = (cntr: UserAccessor) => async (ctx: Koa.Context, fail: Koa.Next, next: Koa.Next) => {
	const accessPayload = ctx.cookies.get(accessTokenName);
	const setting = get_setting();
	const secretKey = setting.jwt_secretkey;

	if (!accessPayload) {
		return await checkRefreshAndUpdate();
	}

	try {
		const payload = await verifyToken(accessPayload, secretKey);
		if (isUserState(payload)) {
			ctx.state.user = payload;
			return await next();
		}
		console.error("Invalid token detected");
		throw new Error("Token form invalid");
	} catch (error) {
		if (error instanceof TokenExpiredError) {
			return await checkRefreshAndUpdate();
		}
		throw error;
	}

	async function checkRefreshAndUpdate() {
		const refreshPayload = ctx.cookies.get(refreshTokenName);
		if (!refreshPayload) {
			return await fail(); // Refresh token doesn't exist
		}

		try {
			const payload = await verifyToken(refreshPayload, secretKey);
			if (isRefreshToken(payload)) {
				const user = await cntr.findUser(payload.username);
				if (!user) return await fail(); // User does not exist

				const permissions = await user.get_permissions();
				const newAccessToken = await createAccessToken({
					username: user.username,
					permission: permissions,
				}, secretKey);

				setToken(ctx, accessTokenName, newAccessToken, accessExpiredTime);
				ctx.state.user = { username: payload.username, permission: permissions };
			} else {
				console.error("Invalid token detected");
				throw new Error("Token form invalid");
			}
		} catch (error) {
			if (error instanceof TokenExpiredError) {
				// Refresh token is expired
				return await fail();
			}
			throw error;
		}

		return await next();
	}
};
export const createRefreshTokenMiddleware = (cntr: UserAccessor) => async (ctx: Koa.Context, next: Koa.Next) => {
	const handler = refreshTokenHandler(cntr);
	await handler(ctx, fail, success);
	async function fail() {
		const user = ctx.state.user as PayloadInfo;
		ctx.body = {
			refresh: false,
			...user,
		};
		ctx.type = "json";
	}
	async function success() {
		const user = ctx.state.user as PayloadInfo;
		ctx.body = {
			...user,
			refresh: true,
			refreshExpired: Math.floor(Date.now() / 1000 + accessExpiredTime),
		};
		ctx.type = "json";
	}
};



export const resetPasswordMiddleware = (cntr: UserAccessor) => async (ctx: Koa.Context, next: Koa.Next) => {
	const body = ctx.request.body;
	const {
		username,
		oldpassword,
		newpassword,
	} = LoginResetRequestSchema.parse(body);
	const user = await cntr.findUser(username);
	if (user === undefined) {
		return sendError(403, "not authorized");
	}
	if (!user.password.check_password(oldpassword)) {
		return sendError(403, "not authorized");
	}
	user.reset_password(newpassword);
	ctx.body = { ok: true };
	ctx.type = "json";
};

export function createLoginRouter(userController: UserAccessor) {
	const router = new Router();
	router.post("/login", createLoginMiddleware(userController));
	router.post("/logout", LogoutMiddleware);
	router.post("/refresh", createRefreshTokenMiddleware(userController));
	router.post("/reset", resetPasswordMiddleware(userController));
	return router;
}

export const getAdmin = async (cntr: UserAccessor) => {
	const admin = await cntr.findUser("admin");
	if (admin === undefined) {
		throw new Error("initial process failed!"); // ???
	}
	return admin;
};

export const isAdminFirst = (admin: IUser) => {
	return admin.password.hash === "unchecked" && admin.password.salt === "unchecked";
};
